#ifndef _FLE_LIST_H_
#define _FLE_LIST_H_

// Backward compatible defines
#define FLE_List FLE_PtrList
#define FLE_Map FLE_ValueMap
#define FLE_Stack FLE_ValueStack

template <class  T> class FLE_PtrList
{
private:
	class Item {
	public:			
		Item(T *t) 	: next(0), prev(0) { stor = t; }
		Item() 			: next(0), prev(0) { stor = 0; }
		Item *next, *prev;
		T *stor;
	};

	bool autodel_;
	int items;
	Item *First, *Last, *Current;
	
	void copy(const FLE_PtrList<T> &list) {
		clear();
  		First = 0;
  		if(list.First) {
  			items = 1;
  			First = new Item(list.First->stor);
  			Item *oldI = list.First;
  			Item *newI = First;  		
  			while(oldI = oldI->next) {
  				items++;
  				newI->next = new Item(oldI->stor);
  				newI = newI->next;
  			}
  		}
	}
	
public:
	FLE_PtrList() : items(0), First(0), Last(0), Current(0) { autodel_=false; }
	FLE_PtrList(const FLE_PtrList<T> &list) { autodel_=false; copy(list); }
	~FLE_PtrList() { clear(); }
	
	void begin() { Current = 0; }
	void end()	 { Current = 0; }
	T *first() 	 { Current = First; return First ? First->stor : 0; }
	T *last() 	 { Current = Last;  return Last ? Last->stor : 0; 	}
	T *current() { return Current ? Current->stor : 0; }
	
	//////////////////////
	// Operators

	// Assing:
	FLE_PtrList<T>& operator = (const FLE_PtrList<T> &list) { if(this==&list) return *this; copy(list); return *this; }
	//Compare:
	bool operator!= (const FLE_PtrList<T> &list ) { return !( *this == list ); }
	bool operator== (const FLE_PtrList<T> &list ) {
		if( count() != list.count() ) return false;
		Item *it = First;
		Item *it2= list.First;
		for(; it!=0; it=it->next, it2=it2->next )
			if ( !(it->stor==it2->stor) ) return false;
		return true;
    }
	// Returns item at 'index', index starts from 0!
	inline T *operator[](int index) {
		return item(index);
	}

	//////////////////	

	inline T *item(int index) {
		if(index < 0 || index >= items)
			return 0;
		Item *it = First;		
		if(index != 0)
			for(int a = 0; a < index; a++)
				it = it->next;
		Current = it;
		return it ? it->stor : 0;
	}

	inline T *item(T *it) {		
		if(!it) return;					
		if(it == First->stor) {			
			Current = First;	
			return Current->stor;
		} else if(it == Last->stor) {			
			Current = Last;
			return Current->stor;
		}
		Current = First;		
		while(Current = Current->next) {
			if(Current->stor == it) {
				break;
			}
		}		
		return Current->stor;
	}
	
	inline void remove(T *it) {		
		if(!it) return;					
		if(it == First->stor) {			
			--items;
			Item *save_next = First->next;
			delete First;			
			First = save_next;
			Current = First;
			return;
		} else if(it == Last->stor) {			
			--items;
			Item *save_prev = Last->prev;
			delete Last;
			Last = save_prev;
			Last->next = 0;
			Current = First;
			return;
		}				

		Item *r = 0;
		Current = First;
		First = Last = Current;
		while(Current = Current->next) {
			if(Current->stor != it) {
				Last->next = Current;
				Current->prev = Last;	
				Last = Current;
			} else {
				--items;
				r = Current;
			}
		}
		if(r) delete r;
		Current = First;
	}
	
	inline void remove(int index) {
		if(index < 0 || index >= items)
			return;
		
		--items;		
		Item *it = First;				
		if(index != 0)
			for(int a = 0; a < index; a++)
				it = it->next;
			
		if(it == First) {			
			First = First->next;
			delete it;
			return;
		} else if(it == Last) {
			Last = Last->prev;
			delete it;
			return;
		}
				
		Item *r;
		Current = First;
		First = Last = Current;
		while(Current = Current->next) {
			if(Current != it) {
				Last->next = Current;
				Current->prev = Last;	
				Last = Current;
			}
			else
				r = Current;
		}
		delete r;	
	}
	
	T *next() {
		if(Current) {
			Current = Current->next;
			if(!Current) return 0;			
		} else
			return first();
		return Current ? Current->stor : 0;
	}
	
	T *prev() {
		if(Current) {
			Current = Current->prev;
			if(!Current) return 0;			
		} else
			return last();
		return Current ? Current->stor : 0;
	}
	
	inline void prepend(T *item) {
		Item *temp = new Item(item);
		items++;
		
		if(!First) {			
			temp->prev = 0;
			temp->next = 0;		
			First = Last = temp;
			return;
		}
		First->prev = temp;
		temp->next = First;
		temp->prev = 0;
		First = temp;		
	}

	inline void append(T *item) {
		Item *temp = new Item(item);
		items++;
		
		if(!First) {			
			temp->prev = 0;
			temp->next = 0;		
			First = Last = temp;
			return;
		}
		Last->next = temp;
		temp->prev = Last;	
		temp->next = 0;			
		Last = temp;
	}
	
	inline void clear() {
		items = 0;
		Item *temp;
		while(First) {
			temp = First->next;
			if(autodel_ && First->stor) delete First->stor;
			delete First;
			First = temp;			
		}
	}
	
	inline const int  count() { return items; }
	inline const bool empty() { return (count()==0); }

	inline bool auto_delete() { return autodel_; }
	inline void auto_delete(bool d) { autodel_ = d; }
};




template <class T> class FLE_ValueListItem {
public:			
	FLE_ValueListItem( const T &t)	: next(0), prev(0), stor(t) {  }
	FLE_ValueListItem()				: next(0), prev(0) { }
	
	FLE_ValueListItem<T> *next;
	FLE_ValueListItem<T> *prev;
	
	T stor;
};

template <class  T> class FLE_ValueListIterator
{
public:
	FLE_ValueListIterator() : item(0) { }
	FLE_ValueListIterator(FLE_ValueListItem<T> *i) : item(i) { }

	FLE_ValueListItem<T> *item;

    bool operator==( const FLE_ValueListIterator<T>& it ) { return item == it.item; }
    bool operator!=( const FLE_ValueListIterator<T>& it ) { return item != it.item; }
    const T& operator*() { return item->stor; }
    FLE_ValueListIterator<T>& operator++() {
		item = item->next;
		return *this;
    }
    FLE_ValueListIterator<T> operator++(int) {
		FLE_ValueListIterator<T> tmp = *this;
		item = item->next;
		return tmp;
    }
    FLE_ValueListIterator<T>& operator--() {
		item = item->prev;
		return *this;
    }
    FLE_ValueListIterator<T> operator--(int) {
		FLE_ValueListIterator<T> tmp = *this;
		item = item->prev;
		return tmp;
    }
};

template <class T> class FLE_ValueList
{
public:
	typedef FLE_ValueListIterator<T> Iterator;
	typedef FLE_ValueListItem<T> Item;
	
protected:	
	int items;
	Item *item_node;
		
	virtual void copy(const FLE_ValueList<T> &list) {
		clear();
  		if(list.items>0) {
			Iterator first( list.item_node->next );
			Iterator last( list.item_node );
			Iterator it( item_node );
			while( first != last) {

				Item *p = new Item(first.item->stor);
				p->next = it.item;
				p->prev = it.item->prev;
				it.item->prev->next = p;
				it.item->prev = p;
				items++;
				
				++first;
			}
  		}
	}

public:	
	FLE_ValueList() : items(0) { item_node = new Item; item_node->next=item_node->prev=item_node; }
	FLE_ValueList(FLE_ValueList<T> &list) { item_node = new Item; item_node->next=item_node->prev=item_node; copy(list); }
	~FLE_ValueList() { clear(); }
	
	//////////////////////
	// Operators

	// Assing:
	FLE_ValueList<T>& operator = (const FLE_ValueList<T> &list) { if(this==&list) return *this; copy(list); return *this; }
	//Compare:
	bool operator!= (const FLE_ValueList<T> &list ) { return !( *this == list ); }
	bool operator== (const FLE_ValueList<T> &list ) {
		if( count() != list.count() ) return false;
		Iterator first = Iterator( item_node->next );
		Iterator first2 = Iterator( item_node->next );
		Iterator end = Iterator( item_node );
		for( ; first != end; ++first, ++first2 )
			if ( !( *first == *first2 ) ) return false;
		return true;
    }
	// Returns item at 'index', index starts from 0!
	inline T operator[](int index) {
		return item(index);
	}

	//////////////////

	inline Iterator begin() { return Iterator(item_node->next); }
	inline Iterator end()	 { return Iterator(item_node); }
	inline Iterator last_iter() { return Iterator(item_node->prev); }

	inline T first() 	 { return item_node->next ? item_node->next->stor : T(); }
	inline T last() 	 { return item_node->prev ? item_node->prev->stor : T(); }
	
	// Returns item at 'index', index starts from 0!
	inline T item(int index) {
		if(index < 0 || index >= items) return T();
		Item *p = item_node->next;
		for(int n = 0; n < index; n++) p = p->next;
		return p ? p->stor : T();
	}

	// Return count of items found
	inline int contains(T &item) {
		int result = 0;
		Iterator first = Iterator( item_node->next );
		Iterator last = Iterator( item_node );
		while( first != last) {
			if ( *first == x )
				++result;
			++first;
		}
		return result;
    }

	inline void remove(int index) { remove(at(index)); }

	//Return number of deleted items
	inline int remove(T &it) {
		int result = 0;
		Iterator first = Iterator(item_node->next);
		Iterator last = Iterator(item_node);
		while(first != last) {
			if(*first == it) {
				first = remove(first);
			++result;
	    } else
			++first;
		}
		return result;
	}	

	// Returns Iterator at 'index', index starts from 0!
	inline Iterator	at(int index) {
		if(index < 0 || index >= items) return Iterator();
		Item *p = item_node->next;
		for(int n = 0; n < index; ++n) p = p->next;
		return Iterator(p);
	}

	inline Iterator remove( Iterator it ) {
		Item *next = it.item->next;
		Item *prev = it.item->prev;
		prev->next = next;
		next->prev = prev;
		delete it.item;
		items--;
		return Iterator(next);
    }
	
	inline Iterator prepend(T &item) {	return insert(begin(), item); }
	inline Iterator append(T &item)  {	return insert(end(), item);   }
	
	// Inserts item T, after index pos
	inline Iterator insert(int pos, T &item) { return insert(at(pos), item); }
	// Inserts item T, after Iterator it
	inline Iterator insert(Iterator it, T &item) {
		Item *p = new Item(item);
		p->next = it.item;
		p->prev = it.item->prev;
		it.item->prev->next = p;
		it.item->prev = p;
		items++;
		return p;
	}
	
	inline void clear() {
		items = 0;
		Item *p = item_node->next;
		while(p != item_node) {
			Item *next = p->next;
			delete p;
			p = next;
		}
		item_node->next = item_node->prev = item_node;
	}
	
	inline const int count() { return items; }
	inline const bool empty() { return (count()==0); }
};


template <class T> class FLE_ValueStack : public FLE_ValueList<T>
{
public:
	FLE_ValueStack() { }

	void push(T &item) { append(item); }	
	T pop() {
		T elem(last());
		if(!empty()) remove(last_iter());
		return elem;
	}
	T top() { return last(); }
};

template <class KEY, class VALUE> class FLE_MapPair {	
public:
	FLE_MapPair(const KEY &_key, const VALUE &_value) : key(_key), value(_value) { }	
	FLE_MapPair(const FLE_MapPair<KEY, VALUE> &m) { key=m.key; value=m.value; }
	FLE_MapPair() { }

	FLE_MapPair<KEY, VALUE>& operator = (const FLE_MapPair<KEY, VALUE> &m) { key=m.key; value=m.value; return *this; }

	KEY key;
	VALUE value;
};

template <class KEY, class VALUE> class FLE_ValueMap : public FLE_ValueList< FLE_MapPair<KEY, VALUE> >
{
public:
	typedef FLE_MapPair<KEY, VALUE> MapPair;

	FLE_ValueMap() { }
	~FLE_ValueMap() { }

	VALUE &value_end() { return item_node->stor.value; }

	void add(KEY &key, VALUE &value) { append(key, value); }
	void append(KEY &key, VALUE &value) {
   		MapPair pair(key, value);
   		append(pair);
	}
	void append(MapPair &pair) {   		
   		FLE_ValueList<MapPair>::append(pair);
	}

	VALUE find(KEY key) { return __get(key, false); }
	VALUE &operator[](KEY key) { KEY tmp(key); return __get(tmp, true); }
	void set(KEY key, VALUE value) { __set(key, value); }

private:
	VALUE &__get(KEY &key, bool add) {				
		Iterator iter = begin();		
		while(iter != end()) {
			if(iter.item->stor.key == key)
				return iter.item->stor.value;
			++iter;
		}
		if(add) {
			VALUE v;
			MapPair val(key, v);
   			Iterator it = FLE_ValueList<MapPair>::append(val);
			return it.item->stor.value;
		}
		return item_node->stor.value;
	}
	
	void __set(KEY &key, VALUE &value) {
		Iterator iter = begin();		
		while(iter != end()) {
			if(iter.item->stor.key == key)
				iter.item->stor.value = value;
			++iter;
		}
		append(key, value);
	}
};

template <class KEY, class VALUE> inline FLE_MapPair<KEY, VALUE> FLE_MakePair(const KEY &key, const VALUE &val) {
	return FLE_MapPair<KEY, VALUE>(key, val);
}

#endif

